/* fptN20.c (release mode) */
/* Mining N-most interesting large itemsets */
/* array of linked lists for support[], largeItemset[][][] and numLarge[] */
/* redo mining if there is not enough k-itemsets */
/* modifications: 1, 6, 7, 8, 4, 9, 13, 14 */
/* N21.c + N16.c */

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#ifndef _WIN32
#include <sys/resource.h>
#include <sys/times.h>
#include <unistd.h>
#endif

#include "trans.h"
#include "fptN20.h"
		

/***** Global Variables *****/
int startHeader;
int fp_round=0;

LargeItemPtr *largeItemset;
int *numLarge;
int *support1;
int *largeItem1;

int *limit;
int *thresholdK;

FPTreeNode root=NULL;
FPTreeNode *headerTableLink;
int headerTableSize;

int N;
int expectedK;
int realK;
int threshold;
int oldThreshold;
int numItem;
int numTrans;
float decreaseRate;
char dataFile[300];
char treeFile[300];
char outFile[300];

int *arrThresholdK;


void destroyTree(FPTreeNode node, int level)
{
 childLink temp1, temp2;

 if (node == NULL) return;

 temp1 = node->children;
 while(temp1 != NULL) {
	temp2 = temp1->next;
	destroyTree(temp1->node, level+1);
	free(temp1);
	temp1 = temp2;
 }

 free(node);

 return;
}

void destroy()
{
 LargeItemPtr aLargeItemset; 
 int i;

 for (i=0; i < realK; i++) {
	aLargeItemset = largeItemset[i];
	while (aLargeItemset != NULL) {
		largeItemset[i] = largeItemset[i]->next;
		free(aLargeItemset->itemset);
		free(aLargeItemset);
		aLargeItemset = largeItemset[i];
	}
 }
 free(largeItemset);

 free(numLarge);
 free(support1);
 free(largeItem1);
 free(limit);
 destroyTree(root, 1);
 //printf("hello\n");
 free(headerTableLink);
 //printf("hello\n");
 free(thresholdK);

 return;
}


/* swap x th element and i th element */
void swap2(int *support, int x, int i)
{ 
 int temp; 

 temp = support[x];
 support[x] = support[i];
 support[i] = temp;
 
 return;
}

/* Quick sort support[] and the corresponding itemset[] in descending order.
 * Parameters:
 *      support[]	-> array to be sorted
 *      itemset[]	-> array to be sorted
 *      low		-> lower bound index of the array to be sorted
 *      high		-> upper bound index of the array to be sorted
 *      size		-> size of the array
 *	length		-> length of an itemset
 */
void q_sort2(int *support, int low,int high, int size)
{
 int pass;
 int highptr=high++;     /* highptr records the last element */
 /* the first element in list is always served as the pivot */
 int pivot=low;

 if(low>=highptr) return;
 do {
	/* find out 1st element value larger than the pivot's */
	pass=1;
	while(pass==1) {
		if(++low<size) {
			if(support[low] > support[pivot])
				pass=1;
			else pass=0;
		} else pass=0;
	} 

	/* find out 1st element value smaller than the pivot's */ 
	pass=1; 
	while(pass==1) {
		if(high-->0) { 
			if(support[high] < support[pivot]) 
				pass=1;
			else pass=0; 
		} else pass=0; 
	}

	/* swap elements pointed by low pointer & high pointer */
	if(low<high)
		swap2(support, low, high);
 } while(low<=high);

 swap2(support, pivot, high);

 /* divide list into two for further sorting */ 
 q_sort2(support, pivot, high-1, size); 
 q_sort2(support, high+1, highptr, size);
 
 return;
}

void visitTree(FPTreeNode T, int level)
{
 childLink child;

 if (level == realK-1) {
 	child = T->children; 
 	while (child != NULL) {
		arrThresholdK[N] = child->node->count;		
		q_sort2(arrThresholdK, 0, N, N+1);

		child = child->next;
	}

	return;
 }

 child = T->children;
 while (child != NULL) {
	visitTree(child->node, level+1);
	child = child->next;
 }

 return;
}


void initThresholdK()
{
 int i;
 int level=0; 

 thresholdK = (int *) malloc (sizeof(int) * (realK+1));
 if (thresholdK == NULL) {
	printf("out of memory\n");
	exit(1);
 }

 arrThresholdK = (int *) malloc (sizeof(int) * (N+1));
 if (arrThresholdK == NULL) {
	printf("out of memory\n");
	exit(1);
 }

 threshold = 1;
 //threshold = 5;
 for (i=0; i <= N; i++) {
	arrThresholdK[i] = threshold;
 }

 visitTree(root, level);

 for (i=1; i < realK; i++) {
	thresholdK[i] = arrThresholdK[N-1];
 }

 free(arrThresholdK);
 
 threshold = thresholdK[realK-1];

 return;
}


/* swap x th element and i th element */
void swap(int *support, int *itemset, int x, int i)
{ 
 int temp; 

 temp = support[x];
 support[x] = support[i];
 support[i] = temp;
 temp = itemset[x];
 itemset[x] = itemset[i];
 itemset[i] = temp;
 
 return;
}

/* Quick sort support[] and the corresponding itemset[] in descending order.
 * Parameters:
 *      support[]	-> array to be sorted
 *      itemset[]	-> array to be sorted
 *      low		-> lower bound index of the array to be sorted
 *      high		-> upper bound index of the array to be sorted
 *      size		-> size of the array
 *	length		-> length of an itemset
 */
void q_sortD(int *support, int *itemset, int low,int high, int size)
{
 int pass;
 int highptr=high++;     /* highptr records the last element */
 /* the first element in list is always served as the pivot */
 int pivot=low;

 if(low>=highptr) return;
 do {
	/* find out 1st element value larger than the pivot's */
	pass=1;
	while(pass==1) {
		if(++low<size) {
			if(support[low] > support[pivot])
				pass=1;
			else pass=0;
		} else pass=0;
	} 

	/* find out 1st element value smaller than the pivot's */ 
	pass=1; 
	while(pass==1) {
		if(high-->0) { 
			if(support[high] < support[pivot]) 
				pass=1;
			else pass=0; 
		} else pass=0; 
	}

	/* swap elements pointed by low pointer & high pointer */
	if(low<high)
		swap(support, itemset, low, high);
 } while(low<=high);

 swap(support, itemset, pivot, high);

 /* divide list into two for further sorting */ 
 q_sortD(support, itemset, pivot, high-1, size); 
 q_sortD(support, itemset, high+1, highptr, size);
 
 return;
}


/* Quick sort indexList[] and freqItemP[] in ascending order.
 * Parameters:
 *      indexList[]     -> array to be sorted
 *      freqItemP[]     -> array to be sorted
 *      low             -> lower bound index of the array to be sorted
 *      high            -> upper bound index of the array to be sorted
 *      size            -> size of the array
 *      length          -> length of an itemset
 */
void q_sortA(int *indexList, int *freqItemP, int low, int high, int size)
{
 int pass;
 int highptr=high++;     /* highptr records the last element */
 /* the first element in list is always served as the pivot */
 int pivot=low;

 if(low>=highptr) return;
 do {
        /* find out 1st element value smaller than the pivot's */
        pass=1;
        while(pass==1) {
                if(++low<size) {
                        if(indexList[low] < indexList[pivot])
                                pass=1;
                        else pass=0;
                } else pass=0;
        }

        /* find out 1st element value larger than the pivot's */
        pass=1;
        while(pass==1) {
                if(high-->0) {
                        if(indexList[high] > indexList[pivot])
                                pass=1;
                        else pass=0;
                } else pass=0;
        }

        /* swap elements pointed by low pointer & high pointer */
        if(low<high)
                swap(indexList, freqItemP, low, high);
 } while(low<=high);

 swap(indexList, freqItemP, pivot, high);

 /* divide list into two for further sorting */
 q_sortA(indexList, freqItemP, pivot, high-1, size);
 q_sortA(indexList, freqItemP, high+1, highptr, size);

 return;
}


void addToLargeList(int *pattern, int patternSupport, int index)
{
 LargeItemPtr aLargeItemset;
 LargeItemPtr aNode, previous=NULL;
 int i;

 aLargeItemset = (LargeItemPtr) malloc (sizeof(ItemsetNode));
 if (aLargeItemset == NULL) {
	printf("out of memory\n");
	exit(1);
 }
 aLargeItemset->itemset = (int *) malloc (sizeof(int) * (index+1));
 if (aLargeItemset->itemset == NULL) {
	printf("out of memory\n");
	exit(1);
 }

 aLargeItemset->support = patternSupport;

 for (i=0; i <= index; i++) {
	aLargeItemset->itemset[i] = pattern[i];
 }

 aLargeItemset->next = NULL;

 aNode = largeItemset[index];
 if (aNode == NULL) {
	largeItemset[index] = aLargeItemset;
 } else {
 	while ((aNode != NULL) && (aNode->support > patternSupport)) {
		previous = aNode;
		aNode = aNode->next;
 	}

 	if (previous != NULL) {
		previous->next = aLargeItemset;
		aLargeItemset->next = aNode;
	} else {
		aLargeItemset->next = largeItemset[index];
		largeItemset[index] = aLargeItemset;
	}
 }

 (numLarge[index])++;

 if (numLarge[index] >= limit[index]) {
        aNode = largeItemset[index];
        for (i=1; i < limit[index]; i++)
                aNode = aNode->next;

        if (thresholdK[index] < aNode->support) {
                thresholdK[index] = aNode->support;

		threshold = thresholdK[realK-1];
		for (i = realK-2; i > 0; i--)
			if (thresholdK[i] < threshold)
				threshold = thresholdK[i];
	}
 }
 
 return;
}


void combine(int *itemList, int *support, int start, int itemListSize, int *base, int baseSize)
{
 int *pattern;
 int i, j; 

 if (baseSize >= realK) return;
 if (start == itemListSize) return;

 pattern = (int *) malloc (sizeof(int) * (baseSize+1));
 if (pattern == NULL) {
	printf("out of memory\n");
	exit(1);
 }

 for (j=0; j < baseSize; j++)
	pattern[j+1] = base[j];

 for (i=start; i < itemListSize; i++) {
	pattern[0] = itemList[i];

	if ((baseSize > 0) && (support[i] < oldThreshold)
		&& (support[i] >= thresholdK[baseSize]))
		addToLargeList(pattern , support[i], baseSize);

	//*** Ray added
	if (support[i] >= threshold)
	{
		combine(itemList, support, i+1, itemListSize, pattern, baseSize+1);	
	}
 }

 free(pattern);
 
 return;
}

void insert_tree(int *freqItemP, int *indexList, int count, int ptr, int length, 
			FPTreeNode T, FPTreeNode *headerTableLink, int *path)  
{
 childLink newNode;
 FPTreeNode hNode;
 FPTreeNode hPrevious;
 childLink previous;
 childLink aNode;

 if (ptr == length) return;

 if (T->children == NULL) {
	/* T has no children */

	newNode = (childLink) malloc (sizeof(ChildNode));
	if (newNode == NULL) {
		printf("out of memory\n");
		exit(1);
	}
	newNode->node = (FPTreeNode) malloc (sizeof(FPNode));
	if (newNode->node == NULL) {
		printf("out of memory\n");
		exit(1);
	}

	newNode->node->item = freqItemP[ptr];
	newNode->node->count = count;
	newNode->node->numPath = 1;
	newNode->node->parent = T;
	newNode->node->children = NULL;
	newNode->node->hlink = NULL;
	newNode->next = NULL;
	T->children = newNode;

	hNode = headerTableLink[indexList[ptr]];
	if (hNode == NULL) {
		headerTableLink[indexList[ptr]] = newNode->node;
	} else {
		while (hNode != NULL) {
			hPrevious = hNode;
			hNode = hNode->hlink;
		}

		hPrevious->hlink = newNode->node;
	}

	insert_tree(freqItemP, indexList, count, ptr+1, length, 
				T->children->node, headerTableLink, path);
	T->numPath += *path;

 } else {
	aNode = T->children;
	while ((aNode != NULL) && (aNode->node->item != freqItemP[ptr])) {
		previous = aNode;
		aNode = aNode->next;
	}

	if (aNode == NULL) {
		/* create a new child for T */ 

		newNode = (childLink) malloc (sizeof(ChildNode));
		if (newNode == NULL) {
			printf("out of memory\n");
			exit(1);
		}
		newNode->node = (FPTreeNode) malloc (sizeof(FPNode));
		if (newNode->node == NULL) {
			printf("out of memory\n");
			exit(1);
		}

		newNode->node->item = freqItemP[ptr];
		newNode->node->count = count;
		newNode->node->numPath = 1;
		newNode->node->parent = T;
		newNode->node->children = NULL;
		newNode->node->hlink = NULL;
		newNode->next = NULL;

		previous->next = newNode;
		
		hNode = headerTableLink[indexList[ptr]];
		if (hNode == NULL)
			headerTableLink[indexList[ptr]] = newNode->node;
		else {
			while (hNode != NULL) {
				hPrevious = hNode;
				hNode = hNode->hlink;
			}
			hPrevious->hlink = newNode->node;
		}

		insert_tree(freqItemP, indexList, count, ptr+1, length, 
					newNode->node, headerTableLink, path);
		(*path)++;
		
		T->numPath += *path;

	} else {
		/* match an existing child of T */

		aNode->node->count += count;

		insert_tree(freqItemP, indexList, count, ptr+1, length, 
				aNode->node, headerTableLink, path);
		T->numPath += *path; 
	}
 }


 return;
}


void buildConTree(FPTreeNode *conRoot, FPTreeNode **conHeader, int conHeaderSize, int *conLargeItem,
		int *conLargeItemSupport, FPTreeNode T, FPTreeNode *headerTable, int baseIndex, int headerSize)
{
 FPTreeNode aNode;
 FPTreeNode aNode2;
 int *freqItemP;
 int *indexList;
 int path;
 int count;
 int i;

 //*** Ray added
 int *mapping;
 int itemNo;

 /* build header table */
 *conHeader = (FPTreeNode *) malloc (sizeof(FPTreeNode) * conHeaderSize);
 if (*conHeader == NULL) {
        printf("out of memory\n");
        exit(1);
 }
 for (i=0; i < conHeaderSize; i++)
        (*conHeader)[i] = NULL;

 /* create root of the FP-tree */
 (*conRoot) = (FPTreeNode) malloc (sizeof(FPNode));
 if (*conRoot == NULL) {
        printf("out of memory\n");
        exit(1);
 }

 (*conRoot)->numPath = 1;
 (*conRoot)->parent = NULL;
 (*conRoot)->children = NULL;
 (*conRoot)->hlink = NULL;

 freqItemP = (int *) malloc (sizeof(int) * conHeaderSize);
 if (freqItemP == NULL) {
        printf("out of memory\n");
        exit(1);
 }

 indexList = (int *) malloc (sizeof(int) * conHeaderSize);
 if (indexList == NULL) {
        printf("out of memory\n");
        exit(1);
 }

 //*** Ray added
 mapping = (int *) malloc(sizeof(int)*numItem);
 for (i = 0; i < numItem; i++)
 {
	 mapping[i] = -1;
 }

 for (i = 0; i < conHeaderSize; i++)
 {
	 
	 itemNo = conLargeItem[i];
	 if (itemNo != -1)
	 {
		mapping[itemNo] = i;
	 }
 }

 aNode = headerTable[baseIndex];
 while (aNode != NULL) {
	aNode2 = aNode->parent;
	count = 0;
	while (aNode2 != T) {
		//*** Ray changed
		//for (i=0; i < conHeaderSize; i++)
			i = mapping[aNode2->item];
			//if (aNode2->item == conLargeItem[i]) {
			if (i != -1) {
				freqItemP[count] = aNode2->item;
				indexList[count] = i;
				count++;
				//break;
			}
		aNode2 = aNode2->parent;
	}

	q_sortA(indexList, freqItemP, 0, count-1, count);

	path = 0;
	insert_tree(&(freqItemP[0]), &(indexList[0]), aNode->count, 0, count, 
				*conRoot, *conHeader, &path);

	aNode = aNode->hlink;
 }

 free(freqItemP);
 free(indexList);
 free(mapping);

 return;
}


void FPgrowth(FPTreeNode T, FPTreeNode *headerTableLink, int headerSize, 
			int *baseItems, int baseSize)
{
 int count;
 int i, j;
 int *pattern;
 int patternSupport;
 int conHeaderSize;
 FPTreeNode *conHeader;
 FPTreeNode conRoot;
 FPTreeNode aNode, aNode2;
 int *conLargeItem;
 int *conLargeItemSupport;
 int localStartHeader=startHeader;

 //*** Ray added
 int *mapping;
 int itemNo;


 if (baseSize >= realK) return;
 if (T == NULL) return;

 //*** Ray added
 mapping = (int *) malloc(sizeof(int)*numItem);
 for (i = 0; i < numItem; i++)
 {
	 mapping[i] = -1;
 }
 for (i = 0; i < headerSize; i++)
 {
	 itemNo = headerTableLink[i]->item;
	 mapping[itemNo] = i;
 }

 if (T->numPath == 1) {
	//printf("single path\n");

	conLargeItem = (int *) malloc (sizeof(int) * headerSize);
	conLargeItemSupport = (int *) malloc (sizeof(int) * headerSize);
	//memset(conLargeItem, 0, sizeof(int) * headerSize);
	//memset(conLargeItemSupport, 0, sizeof(int) * headerSize);
	if ((conLargeItem == NULL) || (conLargeItemSupport == NULL)) {
		printf("out of memory\n");
		exit(1);
	}

	count = 0;
	if (T->children != NULL) aNode = T->children->node;
	while (aNode != NULL) {
		conLargeItem[count] = aNode->item;
		conLargeItemSupport[count] = aNode->count;
		count++;
		if (aNode->children != NULL)
			aNode = aNode->children->node;
		else	aNode = NULL;
	}

	combine(conLargeItem, conLargeItemSupport, 0, count, baseItems, baseSize);

	free(conLargeItem);
	free(conLargeItemSupport);

 } else {
	//printf("multiple path\n");

	pattern = (int *) malloc (sizeof(int) * (baseSize + 1));
	if (pattern == NULL) {
		printf("out of memory\n");
		exit(1);
	}

	conLargeItem = (int *) malloc (sizeof(int) * headerSize);
	conLargeItemSupport = (int *) malloc (sizeof(int) * headerSize);
	if ((conLargeItem == NULL) || (conLargeItemSupport == NULL)) {
		printf("out of memory\n");
		exit(1);
	}

	for (j=0; j < baseSize; j++) {
		pattern[j+1] = baseItems[j];
	}

        if (localStartHeader > headerSize)
                localStartHeader = headerSize;

        for (i=localStartHeader-1; i >= 0; i--) {

                pattern[0] = headerTableLink[i]->item;

                aNode = headerTableLink[i];
                patternSupport = 0;

                while (aNode != NULL) {
                        patternSupport += aNode->count;
                        aNode = aNode->hlink;
                }

                if ((baseSize > 0) && (patternSupport < oldThreshold)
                        && (patternSupport >= thresholdK[baseSize]))
                        addToLargeList(pattern, patternSupport, baseSize);

				//*** Ray added
				if (patternSupport >= threshold)
				{
                /* generate conditional pattern base */

                for (j=0; j < headerSize; j++)
                        conLargeItemSupport[j] = 0;

				for (j = 0; j < headerSize; j++)
				{
					conLargeItem[j] = -1;
				}

                aNode = headerTableLink[i];
                conHeaderSize = 0;
                while (aNode != NULL) {
                        aNode2 = aNode->parent;

                        while (aNode2 != T) {
								//*** Ray changed
                                //for (j=0; j < headerSize; j++)
										j = mapping[aNode2->item];
                                        //if (aNode2->item == headerTableLink[j]->item) {
										if (j != -1) {
                                                conLargeItemSupport[j] += aNode->count;

                                                if ((conLargeItemSupport[j] >= threshold) &&
                                                   (conLargeItemSupport[j] - aNode->count <
                                                        threshold)) {
                                                        conLargeItem[j] = aNode2->item;
                                                        conHeaderSize++;
                                                }
                                                //break;
                                        }
                                aNode2 = aNode2->parent;
                        }
                        aNode = aNode->hlink;
                }


                /* generate conditional pattern tree */
                if (conHeaderSize > 0) {
                        q_sortD(conLargeItemSupport, conLargeItem, 0, headerSize-1, headerSize);

                        buildConTree(&conRoot, &conHeader, conHeaderSize, conLargeItem, conLargeItemSupport,
                                        T, headerTableLink, i, headerSize);

                        FPgrowth(conRoot, conHeader, conHeaderSize, pattern, baseSize+1);

                        free(conHeader);
                        destroyTree(conRoot, 1);
                }
				}
        }


        for (i=localStartHeader; i < headerSize; i++) {

		pattern[0] = headerTableLink[i]->item;

		aNode = headerTableLink[i];
		patternSupport = 0;

		while (aNode != NULL) {
			patternSupport += aNode->count;
			aNode = aNode->hlink;
		}


		if (patternSupport < threshold)
			break;

		if ((baseSize > 0) && (patternSupport < oldThreshold)
			&& (patternSupport >= thresholdK[baseSize]))
			addToLargeList(pattern, patternSupport, baseSize);

		//*** Ray added
		if (patternSupport >= threshold)
		{
		/* generate conditional pattern base */

		for (j=0; j < headerSize; j++)
			conLargeItemSupport[j] = 0;

		aNode = headerTableLink[i];
		conHeaderSize = 0;
		while (aNode != NULL) {
			aNode2 = aNode->parent; 

			while (aNode2 != T) {
				//*** Ray changed
				//for (j=0; j < headerSize; j++)
					//if (aNode2->item == headerTableLink[j]->item) {
					j = mapping[aNode2->item];
					if (j != -1) {
						conLargeItemSupport[j] += aNode->count;  

						if ((conLargeItemSupport[j] >= threshold) &&
						   (conLargeItemSupport[j] - aNode->count < 
							threshold)) {
							conLargeItem[j] = aNode2->item;
							conHeaderSize++;	
						}
						//break;
					}
				aNode2 = aNode2->parent;
			}
			aNode = aNode->hlink;
		}


		/* generate conditional pattern tree */
		if (conHeaderSize > 0) {
			q_sortD(conLargeItemSupport, conLargeItem, 0, headerSize-1, headerSize);

			buildConTree(&conRoot, &conHeader, conHeaderSize, conLargeItem, conLargeItemSupport, 
					T, headerTableLink, i, headerSize);

			FPgrowth(conRoot, conHeader, conHeaderSize, pattern, baseSize+1);

 			free(conHeader);
 			destroyTree(conRoot, 1);
		}
		}
	}

	free(pattern);
	free(conLargeItem);
	free(conLargeItemSupport);
 }

 free(mapping);

 return;
}


int findCombination(int n, int m)
{
 int factorial1=1, factorial2=1;
 int i;

 for (i=n; i > (n-m); i--)
	factorial1 *= i;
 for (i=m; i > 1; i--)
	factorial2 *= i;

 return (factorial1 / factorial2);
}

void findLimit()
{
 int i, j;

 for (i=1; i < realK; i++) {
	for (j=i+1; j < realK; j++) {
		limit[i] += findCombination(j, i);
	}

	if (limit[i] > N)
		limit[i] = N;
 }
 return;
}




void pass1(int startIndex, int bucketSize)
{
 int transSize;
 int item;
 int maxSize=0;
 FILE *fp;
 int i, j;

 support1 = (int *) malloc (sizeof(int) * numItem);
 largeItem1 = (int *) malloc (sizeof(int) * numItem);
 if ((support1 == NULL) || (largeItem1 == NULL)) {
	printf("out of memory\n");
	exit(1);
 }

 limit = (int *) malloc (sizeof(int) * numItem);
 if (limit == NULL) {
	printf("out of memory\n");
	exit(1);
 }
 for (i=0; i < numItem; i++)
	limit[i] = 0;

 for (i=0; i < numItem; i++) { 
	support1[i] = 0;
	largeItem1[i] = i;
 }

 /* scan DB to count frequency of each item */

 if ((fp = fopen(dataFile, "r")) == NULL) {
        printf("Can't open data file, %s.\n", dataFile);
        exit(1);
 }


 // no. of skips
 for (i = 0; i < startIndex; i++)
 {
	fscanf(fp, "%d", &transSize);
	if (transSize > maxSize)
		maxSize = transSize;

	//(limit[transSize-1])++;

	for (j=0; j < transSize; j++) {
		fscanf(fp, "%d", &item);
		//(support1[item])++;
	}
 }

 //for (i=0; i < numTrans; i++) {
 for (i=0; i < bucketSize; i++) {
	fscanf(fp, "%d", &transSize);
	if (transSize > maxSize)
		maxSize = transSize;

	(limit[transSize-1])++;

	for (j=0; j < transSize; j++) {
		fscanf(fp, "%d", &item);
		(support1[item])++;
	}
 } 
 fclose(fp);
 
 /* initialize large itemset list and support list */
 realK = expectedK;
 if ((maxSize < expectedK) || (expectedK <= 0))
	realK = maxSize;
 //printf("maxSize = %d\nrealK = %d\n", maxSize, realK);

 largeItemset = (LargeItemPtr *) malloc (sizeof(LargeItemPtr) * realK); 
 numLarge = (int *) malloc (sizeof(int) * realK);

 if ((largeItemset == NULL) || (numLarge == NULL)) { 
	printf("out of memory\n");
	exit(1);
 }

 for (i=0; i < realK; i++)  {
	largeItemset[i] = NULL;
	numLarge[i] = 0;
 }


 /* sort the support in ascending order */
 q_sortD(&(support1[0]), largeItem1, 0, numItem-1, numItem);

 /*
 for (i=0; i < numItem; i++) 
 	printf("%d[%d] ", largeItem1[i], support1[i]);
 printf("\n");
 */

 headerTableSize = N;
 while ((headerTableSize < numItem) && (support1[N-1] == support1[headerTableSize]))
	headerTableSize++;

 numLarge[0] = headerTableSize;

 headerTableSize = numItem;
 while ((headerTableSize -1 >= 0) && (support1[headerTableSize-1] < 1))
	headerTableSize--;

 /*
 if (headerTableSize >  realK)
 	threshold = support1[N-1];
 else {
	headerTableSize = realK;
 	while ((headerTableSize < numItem) && (support1[realK] == support1[headerTableSize]))
		headerTableSize++;

	if (realK < numItem)
 		threshold = support1[realK];
	else	threshold = support1[realK-1];
 }
 */

 //printf("\nheaderTableSize=%d\n", headerTableSize);
 //printf("numLarge[0]=%d\n", numLarge[0]);


 findLimit();

 return;
}



void buildTree(int startIndex, int bucketSize)
{
 int *freqItemP;
 int *indexList;
 int count;
 FILE *fp;
 int transSize;
 int item;
 int i, j, m;
 int path;

 /* build header table */
 headerTableLink = (FPTreeNode *) malloc (sizeof(FPTreeNode) * headerTableSize);
 if (headerTableLink == NULL) {
	printf("out of memory\n");
	exit(1);
 }

 //for (i=0; i < numItem; i++)
 for (i=0; i < headerTableSize; i++)
	headerTableLink[i] = NULL;

 /* create root of the FP-tree */
 root = (FPTreeNode) malloc (sizeof(FPNode));
 if (root == NULL) {
	printf("out of memory\n");
	exit(1);
 }

 root->numPath = 1;
 root->parent = NULL;
 root->children = NULL;
 root->hlink = NULL;

 freqItemP = (int *) malloc (sizeof(int) * numItem);
 if (freqItemP == NULL) {
	printf("out of memory\n");
	exit(1);
 }	

 indexList = (int *) malloc (sizeof(int) * numItem);
 if (indexList == NULL) {
	printf("out of memory\n");
	exit(1);
 }	


 /* scan DB and insert frequent items into the FP-tree */
 if ((fp = fopen(dataFile, "r")) == NULL) {
        printf("Can't open data file, %s.\n", dataFile);
        exit(1);
 }

 // no. of skips
 for (i = 0; i < startIndex; i++)
 {
	fscanf(fp, "%d", &transSize);
	//if (transSize > maxSize)
	//	maxSize = transSize;

	//(limit[transSize-1])++;

	for (j=0; j < transSize; j++) {
		fscanf(fp, "%d", &item);
		//(support1[item])++;
	}
 }


 for (i=0; i < bucketSize; i++) {
	fscanf(fp, "%d", &transSize);
	count = 0;
 	path = 0;
	for (j=0; j < transSize; j++) {
		fscanf(fp, "%d", &item);
		for (m=0; m < headerTableSize; m++) {
			if (item == largeItem1[m]) {
				freqItemP[count] = item;
				indexList[count] = m;
				count++;
				break;
			} 
		}
	}

	q_sortA(indexList, freqItemP, 0, count-1, count);

	insert_tree(&(freqItemP[0]), &(indexList[0]), 1, 0, count, 
				root, headerTableLink, &path);
 } 
 fclose(fp);

 free(freqItemP);
 free(indexList);

 return;
}


void displayResult()
{
 LargeItemPtr aLargeItemset;
 FILE *fp;
 int count;
 int support;
 int i, j;

 if ((fp = fopen(outFile, "w")) == NULL) {
        printf("Can't open data file, %s.\n", outFile);
        exit(1);
 }
 fprintf(fp, "%d\n\n", realK);

 fprintf(fp, "%d\n", numLarge[0]);
 if (numLarge[0] > 0) {
        printf("\nLarge %d-itemsets[support]:\n", 1);
        for (i=0; i < numLarge[0]; i++) {
                printf("%d ", largeItem1[i]);
                fprintf(fp, "%d ", largeItem1[i]);
                printf("[%d]\n", support1[i]);
                fprintf(fp, "  %d\n", support1[i]);
        }
        printf("\n");
        fprintf(fp, "\n");
 }

 for (i=1; i < realK; i++) {
	if (numLarge[i] == 0) break;
	fprintf(fp, "%d\n", numLarge[i]);
	printf("\nLarge %d-itemsets[support]:\n", i+1);
	aLargeItemset = largeItemset[i];
	count = 0;
	while (aLargeItemset != NULL) {
		count++;
		if (count == N)
			support = aLargeItemset->support;
		else if (count > N)
			if (aLargeItemset->support < support)
				break;
		for (j=0; j <= i; j++) {
			printf("%d ", aLargeItemset->itemset[j]);
			fprintf(fp, "%d ", aLargeItemset->itemset[j]);
		}
		printf("[%d]\n", aLargeItemset->support);
		fprintf(fp, "  %d\n", aLargeItemset->support);
		aLargeItemset = aLargeItemset->next;
	}
	printf("\n");
	fprintf(fp, "\n");
 }

 fclose(fp);

 return;
}


entryPtr *returnBOMOResult(int topK, int L, int *noOfTopK)
{ 
 entryPtr *returnValue;
 LargeItemPtr aLargeItemset;
 FILE *fp;
 int count;
 int support;
 int i, j;

 int count2;
 int *itemset;
 int itemsetSize;
 int freq;

 for (i = 0; i < L; i++)
 {
	 noOfTopK[i] = 0;
 }

 if ((fp = fopen(outFile, "w")) == NULL) {
        printf("Can't open data file, %s.\n", outFile);
        exit(1);
 }
 //fprintf(fp, "%d\n\n", realK);

 returnValue = (entryPtr *) malloc(sizeof(entryPtr)*L);
 for (i = 0; i < L; i++)
 {
	 //returnValue[i] = (entryPtr) malloc(sizeof(entry)*topK*1000);
	 returnValue[i] = (entryPtr) malloc(sizeof(entry));
 }
 itemset = (int *) malloc(sizeof(int)*L);

 itemsetSize = 1;
 count2 = 0;
 //fprintf(fp, "%d\n", numLarge[0]);
 if (numLarge[0] > 0) {
        //printf("\nLarge %d-itemsets[support]:\n", 1);
        for (i=0; i < numLarge[0]; i++) {
                //printf("%d ", largeItem1[i]);
                //fprintf(fp, "%d ", largeItem1[i]);
                //printf("[%d]\n", support1[i]);
                //fprintf(fp, "  %d\n", support1[i]);

				itemset[0] = largeItem1[i];
				freq = support1[i];

				addEntry(&(returnValue[itemsetSize-1]), count2, itemset, itemsetSize, freq);
				count2++;
        }
        //printf("\n");
        //fprintf(fp, "\n");
 }

 noOfTopK[itemsetSize-1] = count2;
 
 for (i=1; i < realK; i++) {
	 count2 = 0;
	itemsetSize = i+1;
	if (numLarge[i] == 0) break;
	//fprintf(fp, "%d\n", numLarge[i]);
	//printf("\nLarge %d-itemsets[support]:\n", i+1);
	aLargeItemset = largeItemset[i];
	count = 0;
	while (aLargeItemset != NULL) {
		count++;
		if (count == N)
			support = aLargeItemset->support;
		else if (count > N)
			if (aLargeItemset->support < support)
				break;
		for (j=0; j <= i; j++) {
			//printf("%d ", aLargeItemset->itemset[j]);
			//fprintf(fp, "%d ", aLargeItemset->itemset[j]);
			
				itemset[j] = aLargeItemset->itemset[j];

		}
		//printf("[%d]\n", aLargeItemset->support);
		//fprintf(fp, "  %d\n", aLargeItemset->support);

				freq = aLargeItemset->support;

				addEntry(&(returnValue[itemsetSize-1]), count2, itemset, itemsetSize, freq);
				count2++;

		aLargeItemset = aLargeItemset->next;


	}
	//printf("\n");
	//fprintf(fp, "\n");

	noOfTopK[itemsetSize-1] = count2;
 }

 fclose(fp);
 free(itemset);

 return returnValue;
}


void input(char *configFile, int topK)
{
 FILE *fp;
 float thresholdDecimal;

 if ((fp = fopen(configFile, "r")) == NULL) {
        printf("Can't open config. file, %s.\n", configFile);
        exit(1);
 }

 fscanf(fp, "%d %d %f %d %d %f", &N, &expectedK, &thresholdDecimal, &numItem, &numTrans, &decreaseRate);
 fscanf(fp, "%s %s %s", dataFile, treeFile, outFile);

 N = topK;
 if (N >= numItem)
 {
	 N = numItem;
 }

 fclose(fp);
/*
 printf("N = %d\n", N);
 printf("expectedK = %d\n", expectedK);
 printf("thresholdDecimal = %f (dummy)\n", thresholdDecimal);
 printf("numItem = %d\n", numItem);
 printf("numTrans = %d\n", numTrans);
 printf("decreaseRate = %f\n", decreaseRate);
 printf("dataFile = %s\n", dataFile);
 printf("treeFile = %s\n", treeFile);
 printf("outFile = %s\n\n", outFile);
*/
 threshold = (int) (thresholdDecimal * numTrans);
 //printf("threshold = %d (dummy)\n", threshold);

 oldThreshold = numTrans;

 if (N > numItem) {
	printf("N > numItem\n");
	exit(1);
 }	
 
 return;
}


//void processBOMO(int argc, char *argv[])
entryPtr *processBOMO(char *configFilename, int startIndex, int bucketSize, int topK, int L, int *noOfTopK)
{
#ifndef _WIN32
 float  userTime, sysTime;
 struct rusage myTime1, myTime2;
#endif
 entryPtr *returnValue;

 input(configFilename, topK);
 startHeader = 0;

 /* usage ------------------------------------------*/
 //printf("fptN20.c\n");
 //printf("FP-tree (finding N-most interesting large itemsets)\n");
 //if (argc != 3) {
 //       printf("Usage: fptN20 <config. file> <where to start scanning the header table>\n");
 //       exit(1);
 //}

 /* read input parameters --------------------------*/
 //printf("input\n");
 //input(argv[1]);
 //startHeader = atoi(argv[2]);
#ifndef _WIN32
 getrusage(RUSAGE_SELF,&myTime1);
#endif



 /* pass 1 -----------------------------------------*/
 //printf("\npass1\n");
 pass1(startIndex, bucketSize);

 //*** Ray changed
 if (startHeader == 0)
	startHeader = headerTableSize/2;//+10*expectedK;//2*expectedK; //startHeader = headerTableSize / 2;
 //printf("startHeader=%d\n", startHeader);

 //*** Ray added
 if (startHeader >= numItem)
 {
	 startHeader = numItem-1;
 }

 /* create FP-tree --------------------------*/
 //printf("\nbuildTree\n");
 buildTree(startIndex, bucketSize);

 //{
//	 int abc;
 //scanf("%d", &abc);
 //}
 /* initialize thresholdK -------------------*/
 //printf("\ninitThreshold\n");
 initThresholdK();

 fp_round++;
 //printf("fp_round = %d, threshold = %d\n", fp_round, threshold);


 /* mining frequent patterns ----------------*/
 //printf("\npassK\n");
 FPgrowth(root, headerTableLink, headerTableSize, NULL, 0);



#ifndef _WIN32
 getrusage(RUSAGE_SELF,&myTime2);
#endif

 /* output result of large itemsets ----------------*/
 //printf("\nresult\n");
 //displayResult();

 //returnValue = returnBOMOResult(int topK, int K, int *noOfTopK);
 returnValue = returnBOMOResult(topK, L, noOfTopK);




 /* free memory ------------------------------------*/
 //printf("\ndestroy\n");
 destroy();

#ifndef _WIN32
 /* output running time ----------------------------*/
 //printf("build tree time:\n");
 userTime =
     		((float) (myTime2.ru_utime.tv_sec  - myTime1.ru_utime.tv_sec)) +
       		((float) (myTime2.ru_utime.tv_usec - myTime1.ru_utime.tv_usec)) * 1e-6;
 sysTime =
       		((float) (myTime2.ru_stime.tv_sec  - myTime1.ru_stime.tv_sec)) +
       		((float) (myTime2.ru_stime.tv_usec - myTime1.ru_stime.tv_usec)) * 1e-6;

 //printf("User time : %f seconds\n",userTime);
 //printf("System time : %f seconds\n\n",sysTime);
#endif


 return returnValue;
}

