
#include "fptN20.h"

#include "string.h"

FPTreeNode rootSet=NULL;
FPTreeNode *headerTableLinkSet;
//int headerTableSize;

extern int *numLarge;
extern int *largeItem1;


extern int *support1;

int *fpMap;




/******************************************************************************************
 * Function: destroyTree
 *
 * Description:
 *	Destroy the FPtree rooted by a node and free the allocated memory.
 *	For each tree node being visited, all the child nodes
 *	are destroyed in a recursive manner before the destroy of the node.
 *
 * Invoked from:	
 * 	destroy()
 * 
 * Input Parameter:
 *	node	-> Root of the tree/subtree to be destroyed.
 */
void destroyFPTree(FPTreeNode node)
{
 childLink temp1, temp2;

 if (node == NULL) return;

 temp1 = node->children;
 while(temp1 != NULL) {
	temp2 = temp1->next;
	destroyFPTree(temp1->node);
	free(temp1);
	temp1 = temp2;
 }

 free(node);

 return;
}


void free_fpSet()
{
	free(headerTableLinkSet);
	destroyFPTree(rootSet);
}


// for normal
void doCreateTrans(double *returnValue, int *freqItemP, int *indexList, FPTreeNode node, short *fpMap, int *isChoose, int *profit)
{
	transPtr transaction;

	int noOfContain;
	int count;

	FPTreeNode parent;

	// check whether the node's count is zero
	if (node->numPath == 0)
	{
		return;
	}
	else
	{
		// create trans
		transaction = (transPtr) malloc(sizeof(trans));
		transaction->item = (int *) malloc(sizeof(int)*noOfItem);
		
		noOfContain = node->numPath;
		
		count = 0;
		//node->numPath -= noOfContain;
		transaction->item[count] = node->item;
		count++;
		
		parent = node->parent;
		
		
		while ((parent != NULL) && (parent->parent != NULL))
		{
			//parent->numPath -= noOfContain;
			transaction->item[count] = parent->item;
			count++;
			
			parent = parent->parent;
		}
		
		transaction->noOfItem = count;
		
		// just create a transaction
		

		//fp_doTransaction(returnValue, transaction, freqItemP, indexList, rootSet, headerTableLinkSet, fpMap, isChoose, profit, noOfContain);
		
		
		
		free(transaction->item);
		free(transaction);
	}
}

// 1: contai
// 0: not contain
int isContainOneItem(int itemNo, int *itemset, int itemsetSize, int *isChoose)
{
	int returnValue;

	int i;

	returnValue = 0;
	for (i = 0; i < itemsetSize; i++)
	{
		if (itemset[i] == itemNo)
		{
			returnValue = 1;
			isChoose[i] = 1;
			break;
		}
	}

	return returnValue;
}

int exploreChild(FPTreeNode node, int *itemset, int itemsetSize, int *isChoose)
{
	childLink nextChild;

	int itemNo;

	int value, sum;
	int j;

	int *isChoose2;

	int returnValue;

	isChoose2 = (int *) malloc(sizeof(int)*itemsetSize);

	nextChild = node->children;
	if (nextChild == NULL)
	{
		return 0;
	}
	returnValue = 0;
	while (nextChild != NULL)
	{
		memcpy(isChoose2, isChoose, sizeof(int)*itemsetSize);

		itemNo = nextChild->node->item;

		// checking itemNo with itemset
		value = isContainOneItem(itemNo, itemset, itemsetSize, isChoose2);
	
		if (value == 1)
		{
			// checking whether isChoose2 = itemstSize
			sum = 0;
			for (j = 0; j < itemsetSize; j++)
			{
				sum += isChoose2[j];
			}
			if (sum == itemsetSize)
			{
				//return node->count;
				returnValue += nextChild->node->count;

				// don't need to explore the child
			}
			else
			{
				returnValue += exploreChild(nextChild->node, itemset, itemsetSize, isChoose2);	
			}
		}
		else
		{
			returnValue += exploreChild(nextChild->node, itemset, itemsetSize, isChoose2);	
		}
		nextChild = nextChild->next;
	}

	free(isChoose2);

	return returnValue;
}

int doNode(FPTreeNode node, int *itemset, int itemsetSize, int *isChoose)
{
	int itemNo;
	FPTreeNode parent;

	int value;

	int sum = 0;

	int i;


	// finding upper part
	itemNo = node->item;

	parent = node->parent;

	while ((parent != NULL) && (parent->parent != NULL))
	{
		itemNo = parent->item;

		// checking itemNo with itemset
		value = isContainOneItem(itemNo, itemset, itemsetSize, isChoose);

		parent = parent->parent;
	}

	// checking whether isChoose = itemstSize
	sum = 0;
	for (i = 0; i < itemsetSize; i++)
	{
		sum += isChoose[i];
	}
	if (sum == itemsetSize)
	{
		return node->count;
	}

	// find lower part
	if (node->children == NULL)
	{
		// this node has no children
		// (i.e. this node is leaf)
		
		return 0;	// as sum != itemsetSize		
	}
	else
	{
		// this node has children
		// (need to explore the children first
		//  by using depth-first search)

		// explore children
		return exploreChild(node, itemset, itemsetSize, isChoose);
		


	}	
	
}

void createFPTree(int startIndex, int bucketSize)
{
 int i,j;
 int *freqItemP;	/* Store frequent items of a transaction */
 int *indexList;	/* indexList[i] = the index position in the large 1-item list storing freqItemP[i] */
 int count;		/* Number of frequent items in a transaction */
 int path;		/* Number of new tree paths (i.e. new leaf nodes) created so far */
 FILE *fp;		/* Pointer to the database file */
 int transSize;		/* Transaction size */
 int item;		/* An item in the transaction */


 fpMap = (int *) malloc(sizeof(int)*noOfItem);



//printf("@@@@@@@ 1.1 (%p)\n", returnValue);
 pass1(startIndex, bucketSize);
//printf("@@@@@@@ 1.2 (%p)\n", returnValue);

 	// create a FP-map
	for (i = 0; i < noOfItem; i++)
	{
		for (j = 0; j < noOfItem; j++)
		{
			if (j == largeItem1[i])
			{
				fpMap[j] = (short) i;
				break;
			}
		}
	}	

	/* Create header table */
 headerTableLinkSet = (FPTreeNode *) malloc (sizeof(FPTreeNode) * numLarge[0]);
 if (headerTableLinkSet == NULL) {
	printf("out of memory\n");
	exit(1);
 }
 for (i=0; i < numLarge[0]; i++)
	headerTableLinkSet[i] = NULL;
	
 /* Create root of the FP-tree */
 rootSet = (FPTreeNode) malloc (sizeof(FPNode));
 if (rootSet == NULL) {
	printf("out of memory\n");
	exit(1);
 }

 /* Initialize the root node */
 rootSet->numPath = 1;
 rootSet->parent = NULL;
 rootSet->children = NULL;
 rootSet->hlink = NULL;

 /* Create freqItemP to store frequent items of a transaction */
 freqItemP = (int *) malloc (sizeof(int) * numItem);
 if (freqItemP == NULL) {
	printf("out of memory\n");
	exit(1);
 }	

 indexList = (int *) malloc (sizeof(int) * numItem);
 if (indexList == NULL) {
	printf("out of memory\n");
	exit(1);
 }	
	



 /* scan DB and insert frequent items into the FP-tree */
 if ((fp = fopen(dataFile, "r")) == NULL) {
        printf("Can't open data file, %s.\n", dataFile);
        exit(1);
 }


 // no. of skips
 for (i = 0; i < startIndex; i++)
 {
	fscanf(fp, "%d", &transSize);
	//if (transSize > maxSize)
	//	maxSize = transSize;

	//(limit[transSize-1])++;

	for (j=0; j < transSize; j++) {
		fscanf(fp, "%d", &item);
		//(support1[item])++;
	}
 }

 for (i=0; i < bucketSize; i++) {
 //for (i=0; i < numTrans; i++) {

	//* Read the transaction size 
	fscanf(fp, "%d", &transSize);

	count = 0;
 	path = 0;

	for (j=0; j < transSize; j++) {

		//* Read a transaction item 
		fscanf(fp, "%d", &item);

		//* Store the item to the frequent list, freqItemP, 
		//* if it is a large 1-item.
		// 
/*		for (m=0; m < numLarge[0]; m++) {
			if (item == largeItem1[m]) {
				// Store the item 
				freqItemP[count] = item;
				// Store the position in the large 1-itemset list storing this item 
				indexList[count] = m;
				count++;
				break;
			} 
		}
*/
		// Store the the item
		freqItemP[count] = item;
		// Store the position in the large 1-itemset list storing this item
		indexList[count] = fpMap[item];
		count++;
	}

	//* Sort the items in the frequent item list in ascending order of indexList,
	// * i.e. sort according to the order of the large 1-itemset list
	// 
	q_sortA(indexList, freqItemP, 0, count-1, count);

	// Insert the frequent patterns of this transaction to the FP-tree. 
	insert_tree(&(freqItemP[0]), &(indexList[0]), 1, 0, count, rootSet, headerTableLinkSet, &path);
 } 
 fclose(fp);


 free(freqItemP);
 free(indexList);
}

void freeTree()
{

 free(largeItem1);
 //free(support1);
 free(fpMap);
 free_fpSet();
}
/******************************************************************************************
 * Function: buildTree()
 *
 * Description:
 *	Build the initial FP-tree.
 *
 * Invoked from:	
 *	main()
 *
 * Functions to be invoked:
 *	insert_tree()
 *	q_sortA()
 *
 * Global variables:
 *	root		-> Pointer to the root of this initial FP-tree
 *	headerTableLink	-> Header table for this initial FP-tree
 *
 * Global variables (read only):
 *	numLarge[]	-> Large k-itemsets resulting list for k = 1 to realK
 */
int *buildTreeSet(int **itemset, int noOfItemset, int itemsetSize, int startIndex, int bucketSize)
{
 int *returnValue;

 
 int i, j;//, m;
 

 int itemNo;
 //int flag;

 //double temp;
 //int isExist;

 int freq;
 int minFreq;
 int min_index;
 int min_j;
 FPTreeNode hsearch;

 int *isChoose;

 returnValue = (int *) malloc(sizeof(int)*noOfItemset);
 memset(returnValue, 0, sizeof(int)*noOfItemset);



 if (itemsetSize == 1)
 {
	 for (i = 0; i < noOfItemset; i++)
	 {
		 returnValue[i] = support1[fpMap[itemset[i][0]]];
		 return returnValue;
	 }
 }

//printf("@@@@@@@ 1.2 (%p)\n", returnValue);

 //createFPTree(startIndex, bucketSize);

 // do what we want
 isChoose = (int *) malloc(sizeof(int)*itemsetSize);
 for (i = 0; i < noOfItemset; i++)
 {
	 if (itemsetSize == 1)
	 {
		 returnValue[i] = support1[fpMap[itemset[i][0]]];
	 }
	 else if (itemsetSize > 0)
	 {

		 // 
		 memset(isChoose, 0, sizeof(int)*itemsetSize);

		 // find the min. freq 
		 itemNo = itemset[i][0];
		 minFreq = support1[fpMap[itemNo]]; // support1[fpMap[itemNo]];  // not sure
		 min_index = itemNo;
		 min_j = 0;
		 
		 for (j = 0; j < itemsetSize; j++)
		 {
			 itemNo = itemset[i][j];
			 freq = support1[fpMap[itemNo]]; // support1[fpMap[itemNo]];  // not sure

			 if (freq == minFreq)
			 {
				 if (fpMap[itemNo] > fpMap[min_index])
				 {
					 minFreq = freq;
					 min_index = itemNo;
					 min_j = j;
				 }
			 }
			 else if (freq < minFreq)
			 {
				 minFreq = freq;
				 min_index = itemNo;
				 min_j = j;
			 }
		 
		 }

		 isChoose[min_j] = 1;

//if (i == 15)
//{
//printf(" ############## returnValue[15] : %d (min_j:%d)\n", returnValue[15], min_j);
//}

		 // 
		 hsearch = headerTableLinkSet[fpMap[min_index]];
		 while (hsearch != NULL)
		 {
			 int abc;

			 memset(isChoose, 0, sizeof(int)*itemsetSize);
			 isChoose[min_j] = 1;

			 //
			 //returnValue[i] += doNode(hsearch, itemset[i], itemsetSize, isChoose);
			 abc = doNode(hsearch, itemset[i], itemsetSize, isChoose);
			 returnValue[i]  += abc;

//if (i == 15)
//{
//printf(" (No:%d; abc:%d) ", hsearch->item,abc);
//}
			 hsearch = hsearch->hlink;
		 }
//if (i == 15)
//{
//printf(" ############## returnValue[15] : %d\n", returnValue[15]);
//}
	 }
	 else
	 {
		 // do nothing
	 }
 }

free(isChoose);

 //freeTree();

 return returnValue;
}
