
#include "heap.h"
#include "trans.h"

void heapEntry_insertTid(heapEntryPtr aEntry, int tid, int index)
{
	int no;

	no = aEntry->noOfTid;
	if (no == 0)
	{
		aEntry->tid = (int *) malloc(sizeof(int)*1);
		aEntry->index = (int *) malloc(sizeof(int)*1);

		aEntry->tid[0] = tid;
		aEntry->index[0] = index;

		no++;
		aEntry->noOfTid = no;
	}
	else
	{
		aEntry->tid = (int *) realloc(aEntry->tid, sizeof(int)*(no+1));
		aEntry->index = (int *) realloc(aEntry->index, sizeof(int)*(no+1));

		aEntry->tid[no] = tid;
		aEntry->index[no] = index;

		no++;
		aEntry->noOfTid = no;
	}
}

heapPtr createHeap(int capacity, int *itemset, int sizeOfItemset)
{
	heapPtr returnValue;
	int i;

	returnValue = (heapPtr) malloc(sizeof(heap));

	returnValue->capacity = capacity;
	returnValue->elt = (int *) malloc(sizeof(int)*capacity);
	returnValue->hEntry = (heapEntryPtr) malloc(sizeof(heapEntry)*capacity);
	memset(returnValue->elt, 0, sizeof(int)*capacity);
	memset(returnValue->hEntry, 0, sizeof(heapEntry)*capacity);
	returnValue->noOfElt = 0;

	for (i = 0; i < capacity; i++)
	{
		returnValue->hEntry[i].itemNo = i;
	}

	returnValue->noOfCondItemset = sizeOfItemset;
	if (sizeOfItemset != 0)
	{
		returnValue->condItemset = (int *) malloc(sizeof(int)*sizeOfItemset);
	}
	for (i = 0; i < sizeOfItemset; i++)
	{
		returnValue->condItemset[i] = itemset[i];
	}
	
	return returnValue;
}

void freeHeap(heapPtr aHeap)
{
	int i;

	for (i = 0; i < aHeap->capacity; i++)
	{
		if (aHeap->hEntry[i].noOfTid != 0)
		{
			free(aHeap->hEntry[i].index);
			free(aHeap->hEntry[i].tid);
		}
	}
	if (aHeap->capacity > 0)
	{
		free(aHeap->elt);
		free(aHeap->hEntry);
	}
	if (aHeap->noOfCondItemset > 0)
	{
		free(aHeap->condItemset);
	}
}

void heap_heapEntry_insertItem(heapPtr aHeap, int itemNo, int tid, int index)
{
	heapEntryPtr hEntryPtr;

	hEntryPtr = &(aHeap->hEntry[itemNo]);
	hEntryPtr->itemNo = itemNo;
	
	heapEntry_insertTid(hEntryPtr, tid, index);
}

void heap_constructElt(heapPtr aHeap)
{
	int i;

	int counter;

	counter = 0;
	for (i = 0; i < aHeap->capacity; i++)
	{
		if (aHeap->hEntry[i].noOfTid != 0)
		{
			aHeap->elt[counter] = aHeap->hEntry[i].itemNo;
			counter++;
		}
	}
	aHeap->noOfElt = counter;
}

int heap_removeMin(heapPtr aHeap)
{
	int returnValue;
	int i;

	if (aHeap == NULL)
	{
		return -1;
	}

	if (aHeap->noOfElt <= 0)
	{
		returnValue = -1;
	}
	else
	{
		returnValue = aHeap->elt[0];

		for (i = 1; i < aHeap->noOfElt; i++)
		{
			aHeap->elt[i-1] = aHeap->elt[i];
		}

		aHeap->noOfElt--;
	}

	return returnValue;
}

heapPtr buildDeepHeap(heapPtr aHeap, int itemNo, int capacity, transPtr *data)
{
	heapPtr returnValue;

	int i, j;
	heapEntryPtr aHEntry;
	int tid;
	int succItemNo;

	int *itemset;
	int sizeOfItemset;

	sizeOfItemset = aHeap->noOfCondItemset+1;
	itemset = (int *) malloc(sizeof(int)*sizeOfItemset);
	for (i = 0; i < sizeOfItemset-1; i++)
	{
		itemset[i] = aHeap->condItemset[i];
	}
	itemset[sizeOfItemset-1] = itemNo;

	returnValue = createHeap(capacity, itemset, sizeOfItemset);

	aHEntry = &(aHeap->hEntry[itemNo]);

	for (i = 0; i < aHEntry->noOfTid; i++)
	{
		tid = aHEntry->tid[i];

		for (j = aHEntry->index[i]+1; j < data[tid]->noOfItem; j++)
		{
			succItemNo = data[tid]->item[j];			

			heap_heapEntry_insertItem(returnValue, succItemNo, tid, j);
		}
	}

	heap_constructElt(returnValue);

	free(itemset);

	return returnValue;
}

heapPtr buildFirstHeap(int capacity, transPtr *data, int noOfTrans)
{
	heapPtr returnValue;

	int i, j;

	int *itemset;
	int sizeOfItemset;

	sizeOfItemset = 0;
	itemset = NULL;

	returnValue = createHeap(capacity, itemset, sizeOfItemset);

	for (i = 0; i < noOfTrans; i++)
	{
		for (j = 0; j < data[i]->noOfItem; j++)
		{
			heap_heapEntry_insertItem(returnValue, data[i]->item[j], i, j);
		}
	}

	heap_constructElt(returnValue);

	return returnValue;
}

void printHeap(heapPtr aHeap)
{
	int i, j;
	heapEntryPtr hEntryPtr;

	printf("  Capacity   : %d\n", aHeap->capacity);
	printf("  CondItemset: ");
	for (i = 0; i < aHeap->noOfCondItemset; i++)
	{
		printf("%d ", aHeap->condItemset[i]);
	}
	printf("\n");
	printf("  No. of Elt : %d\n", aHeap->noOfElt);
	printf("  Elt      : ");
	for (i = 0; i < aHeap->noOfElt; i++)
	{
		printf("%d ", aHeap->elt[i]);
	}
	printf("\n");

	printf("  hEntry   :\n");
	for (i = 0; i < aHeap->capacity; i++)
	{
		hEntryPtr = &(aHeap->hEntry[i]);
		printf("     %d. itemNo: %d   noOfTid: %d", i, hEntryPtr->itemNo, hEntryPtr->noOfTid);
		printf("           tid(index): ");
		for (j = 0; j < hEntryPtr->noOfTid; j++)
		{
			printf("%d(%d) ", hEntryPtr->tid[j], hEntryPtr->index[j]);
		}
		printf("\n");
	}
}

int main2()
{
	int noOfTrans;
	transPtr *data;
	int i, j;

	int itemNo;

	int noOfItem;
	int maxTransSize;

	heapPtr *h;

	maxTransSize = 10;
	noOfItem = 6;

	noOfTrans = 4;
	data = (transPtr *) malloc(sizeof(transPtr)*noOfTrans);
	for (i = 0; i < noOfTrans; i++)
	{
		data[i] = (transPtr) malloc(sizeof(trans));
	}

	data[0]->noOfItem = 2;
	data[0]->item = (int *) malloc(sizeof(int)*data[0]->noOfItem);
	data[0]->item[0] = 0;
	data[0]->item[1] = 1;

	data[1]->noOfItem = 2;
	data[1]->item = (int *) malloc(sizeof(int)*data[1]->noOfItem);
	data[1]->item[0] = 2;
	data[1]->item[1] = 3;

	data[2]->noOfItem = 4;
	data[2]->item = (int *) malloc(sizeof(int)*data[2]->noOfItem);
	data[2]->item[0] = 0;
	data[2]->item[1] = 1;
	data[2]->item[2] = 2;
	data[2]->item[3] = 4;
	
	data[3]->noOfItem = 2;
	data[3]->item = (int *) malloc(sizeof(int)*data[3]->noOfItem);
	data[3]->item[0] = 0;
	data[3]->item[1] = 1;

	printf("Transactions : \n");
	maxTransSize = 0;
	for (i = 0; i < noOfTrans; i++)
	{
		if (data[i]->noOfItem > maxTransSize)
		{
			maxTransSize = data[i]->noOfItem;
		}

		printf("  %d. ", i);
		for (j = 0; j < data[i]->noOfItem; j++)
		{
			printf("%d ", data[i]->item[j]);
		}
		printf("\n");
	}

	h = (heapPtr *) malloc(sizeof(heap)*maxTransSize);

	printf("Building first heap...\n");
	h[0] = buildFirstHeap(noOfItem, data, noOfTrans);

	printf("Printing the first heap....\n");
	printHeap(h[0]);

	for (j = 0; j < noOfItem; j++)
	{
		printf("---------------\n");
		for (i = 1; i < maxTransSize; i++)
		{
			
			itemNo = heap_removeMin(h[i-1]);

			if (itemNo != -1)
			{
				printf("Building deep heap (itemNo: %d)...\n", itemNo);
				h[i] = buildDeepHeap(h[i-1], itemNo, noOfItem, data);

				printf("Printing the deep heap (itemNo: %d)....\n", itemNo);
				printHeap(h[i]);

			}
			else
			{
				h[i] = NULL;
			}
		}

		for (i = 1; i < maxTransSize; i++)
		{
			if (h[i] != NULL)
			{
				freeHeap(h[i]);
				free(h[i]);
			}
		}
	}

	freeHeap(h[0]);
	free(h[0]);
	free(h);

	return 0;
}
